/*
  LinEAK - Linux support for Easy Access and Internet Keyboards      
  Copyright (C) 2001,2002 Mark Smulders <Mark@PIRnet.nl> and
  Sheldon Lee-Wen <leewsb@hotmail.com>
  http://lineak.sourceforge.net

  [NOTE]
    The X11/Xkb related functions in this file (initializeX, GetDisplay, commitXKBchanges)
	are based on the work of Anthony Y P Wong <ypwong@ypwong.org>
    I owe him a great deal for inspiring me with HOTKEYS.
    Thanks Anthony!
	
  This program is free software; you can redistribute it and/or modify  
  it under the terms of the GNU General Public License as published by  
  the Free Software Foundation; either version 2, or (at your option)   
  any later version.                                                    
                                                                        
  This program is distributed in the hope that it will be useful,       
  but WITHOUT ANY WARRANTY; without even the implied warranty of        
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
  GNU General Public License for more details.                          
                                                                        
  You should have received a copy of the GNU General Public License     
  along with this program; if not, write to the                         
  Free Software Foundation, Inc., 59 Temple Place - Suite 330,          
  Boston, MA 02111-1307, USA.                                           
*/

/* settings from configure */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_XOSD
#  include <xosd.h>
#  include <sys/time.h>
#endif

/* system header files */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#ifdef HAVE_GETOPT_H
#  define HAVE_GETOPT_LONG
#  include <getopt.h>
#endif
#include <sys/types.h>
#include <dirent.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/XKBlib.h>
#include <X11/extensions/XKBfile.h>
#ifndef X_NOT_STDC_ENV
#include <stdlib.h>
#else
extern char *getenv();
#endif
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/cdrom.h> /* linux specific?!! */

extern int mkdir(); /* in linux/fs.h ... other platforms!? */

#if defined (__FreeBSD__)
# include <machine/soundcard.h>
#else
# if defined (__NetBSD__) || defined (__OpenBSD__)
#  include <soundcard.h>          /* OSS emulation */
#  undef ioctl
# else
/* BSDI, Linux, Solaris */
#  include <sys/soundcard.h>
# endif
#endif

/* lineakd header file */
#include "lineakd.h"

//#include <dmalloc.h>

/**********************************************************************************************/

/* main function */
int main(int argc, char *argv[]) {
  XkbEvent xev;

  /* parse command line arguments */
  if ( !parseargs(argc,argv) )
	cleanexit(1);
  
  /* read the configuration and keyboard definition file */
  parseconffile();
  parsedeffile();

  /* check if a pid file exists and take care of it */
  check_pid_on_startup();
  
  /* everything went according to plan thisfar... some signal-handlers for cleaner exiting */
  signal(SIGTERM, signalexit);
  signal(SIGABRT, signalexit);
  signal(SIGINT, signalexit);
  /* and one so we won't have to wait() for child processes ;) */
  signal(SIGCLD, SIG_IGN);
  /* and for a rehash when we catch SIGHUP */
  signal(SIGHUP,signalHUP);
  
  /* init the cdrom (disable auto eject) */
  if (strcmp(cdromdev,"")!=0) { initCDROM (); }
  /* init sound (store master volume of MIXERDEV) */
  initSound();
  
  /* build our EAKeyboard with the data from deffile and conffile */
  initEAK();
  
  /* we used the data from the files, clean it from memory */
  dcfp_destroy_data_struct((dcfp_data_struct *)confdata);
  dcfp_destroy_data_struct((dcfp_data_struct *)defdata);

  
  /* just free the user def data, the tree gets destroyed above already */
  if (usrdefdata != NULL) free((dcfp_data_struct *)usrdefdata);

  /* initialize XOSD system */
  #ifdef HAVE_XOSD
  osd = init_lxosd(font, color, delay, pos, offset, shadow, lines);
  #endif
  /* init X/Xkb with our EAK keycodes/keysyms */
  initializeX();
 
  /* do we daemonize now? do we? do we? ...ahhh, please? ;-) */
  if (runasdaemon)
	daemon(0,0); // stdin, stdout and stderr to /dev/null when daemon...
  
  /* alright, we're going to loop forever now. only signals can interrupt us. */
  while(1) {
	/* wait for next event */
	XNextEvent(ourdisplay, &xev.core);
	/* should we respond to this event? */
    	if ( xev.type == xkbEventCode+XkbEventCode && xev.any.xkb_type == XkbActionMessage ) {
      		if (verbosemode) printf("Keycode %d pressed\n", xev.message.keycode);
		/* if we are using xosd, output our display here */
                #ifdef HAVE_XOSD
		/* Should we age the display? */
		lxosd_agenscroll();
                #endif
	  	/* find the corresponding key, and execute it's command */
	  	execEAcommand(xev.message.keycode);
	}
	if ( xev.core.type == MappingNotify ) {
		if (!xkbremapped)
			xkbremapped =1;
		else {
			if (verbosemode)
				printf("Reclaiming keyboard map from MappingNotify event\n");
			xkbremapped = 0;
			initializeX();
		}
	} 
        #ifdef HAVE_XOSD
	lxosd_age();
        #endif 
  }
  /* that's all folks.. */
  cleanup(); // just for fun.. we don't ever get here ;-)

  return 0;
}

/**********************************************************************************************/

/* parse .conf file */
void parseconffile(void) {
  char *homedir = NULL;
  const void *temp = NULL;

  /* set (globally) full path of config file ($HOME/.lineak/CONFIGFILE) */
  if (conffilename == NULL) {
    homedir = getenv("HOME");
    conffilename = malloc((strlen(homedir) + strlen(CONFFILE) + 1) * sizeof(char));
    sprintf(conffilename, "%s%s", homedir, CONFFILE);
  }

  /* does the file exist? */
  if (!file_exists(conffilename)) {
	free(conffilename);
	conffilename = malloc((strlen(SYS_CONFFILE) + 1) * sizeof(char));
	sprintf(conffilename, "%s", SYS_CONFFILE);
	if (!file_exists(conffilename)) {
	   /* file does not exist */
           printf("\n*** configuration file %s not found,\n    please run \"lineakd -c TYPE\" to create one for your keyboard.\n\n",conffilename);
	   cleanexit(1);
	}
  }
  /* now that we're sure it exists, parse the contents */
  if ( (confdata = dcfp_parse_file(conffilename)) == NULL ) {
	/* error parsing file */
	printf("*** FATAL ERROR occurred while reading configuration data!\n");
	cleanexit(1);
  }
  /* store keyboard type globally */
  if ((temp = dcfp_get_global_var(confdata, "KeyboardType"))==NULL) {
	printf ("*** Error in config file: KeyboardType not found!\n");
	cleanexit(1);
  } else {
	if (kbtype!=NULL) { free(kbtype); kbtype = NULL; }
	kbtype = (char *)malloc((strlen((char *)temp)+1)*sizeof(char));
	sprintf(kbtype, "%s", (char *)temp);
  }
  /* if not specified from command line, store cdrom device globally, defaults to /dev/cdrom */
  if (cdromdev == NULL) {
    if ((temp = dcfp_get_global_var(confdata, "CdromDevice"))==NULL) {
	  printf ("*** CdromDevice not found in config file, setting default to /dev/cdrom...\n");
	  cdromdev = (char *)malloc(11*sizeof(char));
	  sprintf(cdromdev, "%s", "/dev/cdrom");
    } else {
	  cdromdev = malloc((strlen((char *)temp)+1)*sizeof(char));
	  sprintf(cdromdev, "%s", (char *)temp);
	}
  }
  /* if not specified from command line, store mixer device globally, defaults to /dev/mixer */
  if (mixdev == NULL) {
    if ((temp = (const char *)dcfp_get_global_var(confdata, "MixerDevice"))==NULL) {
	  printf ("*** MixerDevice not found in config file, setting default to /dev/mixer...\n");
	  mixdev = (char *)malloc(11*sizeof(char));
	  sprintf(mixdev, "%s", "/dev/mixer");
    } else {
	  mixdev = malloc((strlen((char *)temp)+1)*sizeof(char));
	  sprintf(mixdev, "%s", (char *)temp);
	}
  }
}

/* create new .conf file in homedir for keyboard 'type' */
void create_new_conf(char *homedir, char *type) {
  FILE *fp;
  char *lineakdir;
  const dcfp_section_struct *kbsect, *keysect;
  dcfp_var_struct *keyptr;

  if (type == NULL) return;
	
  /* get homedir */
  if (homedir == NULL)
    homedir = getenv("HOME");	

  /* check if $HOME/.lineak/ directory exists, if not create it */
  lineakdir = malloc( (strlen(homedir) + 1 + 7 + 1) * sizeof(char));
  sprintf(lineakdir, "%s/.lineak", homedir);
  if (opendir(lineakdir) == NULL) {
    if (mkdir(lineakdir, 0755) == -1) {
      printf("*** FATAL ERROR: unable to create directory %s !\n",lineakdir);
	  cleanexit(1);
    } 
  }
  free(lineakdir);
  
  /* set (globally) full path of config file ($HOME/.lineak/CONFIGFILE) */
  if (conffilename == NULL) {
    conffilename = malloc((strlen(homedir) + strlen(CONFFILE) + 1) * sizeof(char));
    sprintf(conffilename, "%s%s", homedir, CONFFILE);
  }

  /* open the file for write */
  if ((fp = fopen(conffilename, "w")) != NULL) {
    /* no CD-ROM device specified? */
    if ((cdromdev!=NULL) && (*cdromdev=='\0')) {
	  free (cdromdev);
	  cdromdev = NULL;
	}
    if (cdromdev==NULL)
	  printf("*** No CD-ROM device specified, using the default /dev/cdrom\n");
    /* no mixer device specified? */
    if ((mixdev!=NULL) && (*mixdev=='\0')) {
	  free (mixdev);
	  mixdev = NULL;
	}
    if (mixdev==NULL)
	  printf("*** No mixer device specified, using the default /dev/mixer\n");

    /* do we have a valid keyboard type? */
    if (!dcfp_global_section_exist(defdata, type)) {
	  printf("\n*** ERROR: Invalid keyboard type: %s\nTo find out the supported keyboard types, use: 'lineakd -l'\n\n",type);
	  cleanexit(1);
    }
    kbsect = dcfp_get_global_section_by_name(defdata, type);
	printf("\n*** Creating fresh configuration in %s \n      for keyboard type: %s\n\n",conffilename,type);
    fprintf(fp,"%s# LinEAK Configuration file for %s %s\n\n# Global settings\n"
                 "KeyboardType\t\t= %s\n"
                 "CdromDevice\t\t= %s\n"
                 "MixerDevice\t\t= %s\n"
                 "\n# Specific keys of your keyboard\n", \
	        CONF_HEADER, (char *)dcfp_get_var_from_section(kbsect,"brandname"), (char *)dcfp_get_var_from_section(kbsect,"modelname"), \
	        type, cdromdev==NULL?"/dev/cdrom":cdromdev, mixdev==NULL?"/dev/mixer":mixdev);
    if ((keysect = dcfp_get_subsection_by_name(kbsect, "KEYS")) == NULL) {
      printf("*** Error: no keys found in your keyboard section!\n");
    } else {
      keyptr = keysect->globalvars;
      while (keyptr != NULL) {
        fprintf(fp,"%s%s= \n",keyptr->key,(strlen(keyptr->key)>7)?"\t":"\t\t");
	    keyptr = keyptr->next;
	  }
	}
	fprintf(fp,"\n# end lineakd.conf\n");
    fclose(fp);	
	printf("NOTE: Now please edit the file and bind commands to the keys,\n      or use lineakconfig :)\n\n");
  } else {
    printf("*** FATAL ERROR: unable to write configuration file %s !\n",conffilename);
	cleanexit(1);
  }
}

/* parse keyboard definition file */
void parsedeffile(void) {
  char *home=NULL;
  dcfp_section_struct *sectptr=NULL;
  dcfp_var_struct *varptr=NULL;

  /* SET full path of definition files ($sysconfdir/lineakkb.def) */
  if (usrdeffilename == NULL) {
	home = getenv("HOME");
	usrdeffilename = malloc( (strlen(home) + 1 + 7 + 1 + strlen(DEFFILE) +1) * sizeof(char));
	sprintf(usrdeffilename,"%s/.lineak/%s", home, DEFFILE);
	if (file_exists(usrdeffilename)) {
	  /* it exists, parse the contents */
	  if ( (usrdefdata = dcfp_parse_file(usrdeffilename)) == NULL) {
	    /* error parsing file */
	    printf("*** Error occurred while reading definition data from %s!\n", usrdeffilename);
		dcfp_destroy_data_struct((dcfp_data_struct *)usrdefdata);
		usrdefdata = NULL;
 	  }
	}
  }
  if (deffilename == NULL) {
    deffilename = malloc((strlen(CONFDIR) + 1 + strlen(DEFFILE) + 1)*sizeof(char));
    sprintf(deffilename, "%s/%s", CONFDIR, DEFFILE);
  }

  /* does the file exist? */
  if (file_exists(deffilename)) {
	/* it exists, parse the contents */
    if ( (defdata = dcfp_parse_file(deffilename)) == NULL) {
	  /* error parsing file */
	  printf("*** FATAL ERROR occurred while reading definition data from %s!\n", deffilename);
  	  cleanexit(1);
	}
  } else {
	/* file does not exist */
    printf("*** FATAL ERROR: definition file %s not found!\n",deffilename);
	cleanexit(1);
  }
  
  if (usrdefdata != NULL) { //append user def data
    varptr = defdata->globalvars;
	if (varptr==NULL) {
	  ((dcfp_data_struct *)defdata)->globalvars = usrdefdata->globalvars;
	} else {
	  while (varptr->next != NULL) varptr = varptr->next;
	  varptr->next = usrdefdata->globalvars;
	}
    sectptr = defdata->sections;
	if (sectptr == NULL) {
	  ((dcfp_data_struct *)defdata)->sections = usrdefdata->sections;
	} else {
  	  while (sectptr->next != NULL) sectptr = sectptr->next;
	  sectptr->next = usrdefdata->sections;
	}
  }
}

/* check if a file exists */
int file_exists(char *filename) {
  FILE *fp;
  if ((fp = fopen(filename, "r"))!=NULL) {
	fclose (fp);
	return 1;
  } else
	return 0;
}

/* for deleting a file */
void delete_file(char *filename) {
  char *rmfile=NULL;
  if (filename == NULL) return;
  if (file_exists(filename)) {
    rmfile = malloc ( ( 3 + strlen(filename) + 17 + 1) * sizeof(char));
    sprintf(rmfile, "rm %s >>/dev/null 2>&1", filename);
    system(rmfile);
    free(rmfile);
  }
}

/* check if there's already a running lineakd for this $USER */
void check_pid_on_startup(void) {
  FILE *fp;
  int pspid=0;
  pid_t mypid;
  char *home, pidstr[9], *pidtestcmd=NULL, *pidtestfile=NULL, *lineakdir;
  pidstr[0]='\0';

  mypid = getpid();
  home = getenv("HOME");
  pidtestfile = malloc ( ( strlen(home) + strlen(PIDFILE) + 6) * sizeof(char));
  sprintf(pidtestfile, "%s%s.temp", home, PIDFILE);

  /* Make sure the .lineak directory exists */
  lineakdir = malloc( (strlen(home) + 1 + 7 + 1) * sizeof(char));
  sprintf(lineakdir, "%s/.lineak", home);
  if (opendir(lineakdir) == NULL) {
     if (mkdir(lineakdir, 0755) == -1) {
        printf("*** FATAL ERROR: unable to create directory %s !\n",lineakdir);
        cleanexit(1);
     }
  }
  free(lineakdir);
  
  /* get pid from 'ps' and store in pidtestfile */
  sprintf(pidstr, "%d", mypid);
  pidtestcmd = malloc( ( 63 + strlen(pidstr) + 24 + strlen(pidtestfile) + 1) * sizeof(char));
  sprintf(pidtestcmd, "ps -ef | grep \"lineakd \" | grep $USER | grep -v grep | grep -v %d | awk '{ print $2 }' > %s ", mypid, pidtestfile);
  system(pidtestcmd);
  /* get pid from pidtestfile */
  if ((fp = fopen(pidtestfile, "r"))!=NULL) {
    if (fgets(pidstr, 8, fp) != NULL) {
	  if (pidstr[strlen(pidstr)-1] == '\n' )
		pidstr[strlen(pidstr)-1] = '\0';
	  sscanf(pidstr, "%d", &pspid);
	}
	fclose(fp);
    delete_file(pidtestfile);
  } else {
    printf("*** Lost our PID file... I'm confused ...\n");
  }
  /* test the pid from 'ps' ... if its valid we bail out */
  if (pspid != 0) {
	printf("\n*** lineakd already running (pid %d)... exiting...\n\n", pspid);
    free(pidtestcmd);
    free(pidtestfile);
    cleanexit(1);
  }
  free(pidtestcmd);
  free(pidtestfile);
}

/**********************************************************************************************/

/* show supported keyboards */
void showkeyboards (void) {
  dcfp_section_struct *ptr=NULL;
  printf("\nLinEAK v" VERSION " -- supported keyboards:\n\n");
  printf(" [TYPE]\t\t[Full name]\n\n");
  ptr = defdata->sections;
  while (ptr!=NULL) {
	printf(" %s%s%s %s\n",ptr->name,strlen(ptr->name)>=7?"\t":"\t\t",strcasecmp((char *)dcfp_get_var_from_section(ptr,"brandname"),"other")==0?"":(char *)dcfp_get_var_from_section(ptr,"brandname"),(char *)dcfp_get_var_from_section(ptr,"modelname"));
	ptr = ptr->next;
  }
  printf("\n");
}

/* init the EAKeyboard */
int initEAK (void) {
  const dcfp_section_struct *kbsect=NULL, *keysect=NULL;
  dcfp_var_struct *keyptr=NULL;
  EAkey *tmp = NULL;
  const char *tmpvar=NULL;
  int keycount = 0;

  if ((kbsect = dcfp_get_global_section_by_name(defdata, kbtype)) == NULL) {
	printf ("\n*** ERROR: your keyboard (%s) was not found in definition file %s\n\n",kbtype, deffilename);
	cleanexit(1);
  }
  myEAK.EAKtype = kbtype;
  if ((tmpvar = (const char *)dcfp_get_var_from_section(kbsect, "brandname"))==NULL) {
	printf ("\n*** ERROR: no 'brandname' found in the section of your keyboard (%s)\n\n",kbtype);
	cleanexit(1);
  } else {
    myEAK.EAKbrand = malloc((strlen(tmpvar)+1)*sizeof(char));
	sprintf(myEAK.EAKbrand, "%s", tmpvar);
  }
  if ((tmpvar = (const char *)dcfp_get_var_from_section(kbsect, "modelname"))==NULL) {
	printf ("\n*** ERROR: no 'modelname' found in the section of your keyboard (%s)\n\n",kbtype);
	cleanexit(1);
  } else {
    myEAK.EAKmodel = malloc((strlen(tmpvar)+1)*sizeof(char));
	sprintf(myEAK.EAKmodel, "%s", tmpvar);
  }
  if (verbosemode) printf("Configured your keyboard as \"%s\" (%s %s)\n",myEAK.EAKtype,myEAK.EAKbrand,myEAK.EAKmodel);
  if ((keysect = dcfp_get_subsection_by_name(kbsect, "KEYS")) == NULL) {
	printf("\n*** Error: no [KEYS] section found in your keyboard definition (%s)\n\n",kbtype);
	cleanexit(1);
  }
  if ((keyptr = keysect->globalvars) == NULL) {
	printf("\n*** Error: no keys found in the [KEYS] section of your keyboard (%s)\n\n",kbtype);
	cleanexit(1);
  }
  if (keyptr != NULL) {
	myEAK.EAKeylist = tmp = malloc(sizeof(EAkey));
	while(keyptr!=NULL) {
      keycount++;
	  tmp->EAkeycode = *((int *)(keyptr->value));
	  tmp->EAkeysym = 0x1008FF01 + keycount;
	  tmp->EAcommand = NULL;
	  if ( keyptr->key==NULL || *(keyptr->key)=='\0') {
		printf("*** ERROR while fetching keyname for key with keycode %d from configuration file!\n",tmp->EAkeycode);
		cleanexit(1);
	  } else {
		tmp->EAkeyname = malloc((strlen(keyptr->key)+1)*sizeof(char));
		sprintf(tmp->EAkeyname,"%s",keyptr->key);
	  }
	  tmpvar = dcfp_get_global_var(confdata, tmp->EAkeyname);
	  if ( tmpvar==NULL || *tmpvar=='\0') {
		//printf("*** ERROR while fetching command for key '%s' from configuration file!\n",tmp->EAkeyname);
		tmp->EAcommand = malloc(sizeof(char));
		*(tmp->EAcommand) = '\0';
	  } else {
		tmp->EAcommand = malloc((strlen((char *)tmpvar)+1)*sizeof(char));
		sprintf(tmp->EAcommand,"%s",(char *)tmpvar);
	  }
	  if (verbosemode) printf("Configured key \"%s\" with keycode %d, keysym 0x%x and command: %s\n",
		                       tmp->EAkeyname,tmp->EAkeycode,(unsigned int)tmp->EAkeysym,
	                           strcmp(tmp->EAcommand,"")==0?"(none)":tmp->EAcommand);
	  if (keyptr->next != NULL) {
		tmp->next = malloc(sizeof(EAkey));
		tmp = tmp->next;
	  } else {
		tmp->next = NULL;
	  }
	  keyptr = keyptr->next;
	}
  } else {
	myEAK.EAKeylist = NULL;
  }
  return 0;
}

/* run the proper command of the key with keycode 'keycode' */
void execEAcommand(int keycode) {
  EAkey *theKey = NULL;
  char *theCmd = NULL;
  int retval = 0;
  /* lookup the key in our EAKeylist */
  if ((theKey = lookupEAkey(keycode)) == NULL)
	/* oops, not found... how did THAT happen? */
	printf("Key with keycode %d not found!!\n",keycode);
  else {
	if (verbosemode) printf("... that's the \"%s\" key.\n",theKey->EAkeyname);
	/* check if the key has to run a SPECIAL command */
	if (strcmp(theKey->EAcommand,EAK_EJECT)==0) { /* Eject */
          if (strcmp(cdromdev,"")!=0) {
		/** Output the command */
		#ifdef HAVE_XOSD
                lxosd_show ("Ejecting CDRom");
                #endif
		DoEject();
	  }
	} else if (strcmp(theKey->EAcommand,EAK_VOLUP)==0) { /* VolumeUp */
	  retval = AdjustVolume(VOLUP_VALUE);
          #ifdef HAVE_XOSD
	  lxosd_volume(retval);
          #endif
	} else if (strcmp(theKey->EAcommand,EAK_VOLDOWN)==0) { /* VolumeDown */
	  retval = AdjustVolume(VOLDOWN_VALUE);
          #ifdef HAVE_XOSD
	  lxosd_volume(retval);
	  #endif
	} else if (strcmp(theKey->EAcommand,EAK_MUTE)==0) { /* Mute */
	  retval = AdjustVolume(VOLMUTE_VALUE);
          #ifdef HAVE_XOSD
	  lxosd_volume(retval);
          #endif
	} else if (strcmp(theKey->EAcommand,EAK_SLEEP)==0) { /* Sleep (not implemented yet) */
          #ifdef HAVE_XOSD
	  lxosd_show ("Sleep");
          #endif
        } else if (strcmp(theKey->EAcommand,"")==0) { /* no command defined */
	  if (verbosemode) printf("... but it has no command bound to it :(\n");
	} else {
	  /* all specials done, let's go for it... ;) */
	  if (verbosemode) printf("... running \"%s\"\n",theKey->EAcommand);
          #ifdef HAVE_XOSD
	  lxosd_show (theKey->EAcommand);
          #endif
	  if (!fork()) {
		/* child process that tries to run the command, and then exits */
		theCmd = malloc((strlen(theKey->EAcommand)+3)*sizeof(char));
		sprintf(theCmd, "%s &", theKey->EAcommand);
		system(theCmd);
		free(theCmd);
		cleanexit(1);
	  }
	  /* NOTE: no, we don't wait() for the child. we just ignore the SIGCLD signal */
	}
  }
  //printf ("retval is %d\n",retval);
  retval = 0;
}

/* find the key with keycode 'keycode' in myEAK.EAKeylist */
EAkey *lookupEAkey(int keycode) {
  EAkey *tmp = NULL;
  for (tmp = myEAK.EAKeylist; tmp!=NULL; tmp=tmp->next)
	if (tmp->EAkeycode == keycode)
	  break;
  return tmp;	  
}

/**********************************************************************************************/

/* init CD-ROM: disable SW eject */
void initCDROM (void) {
  int fp;
  if (verbosemode) printf("CD-ROM init, using \"%s\" as the CD-ROM device\n",cdromdev);
  /* do we have a CD-ROM device configured?  (REDUNDANT extra check) */
  if (strcmp(cdromdev,"")!=0) {
	/* try to open the device .. */
    if ( (fp = open(cdromdev, O_RDONLY | O_NONBLOCK)) != -1 ) {
	  /* tell the drive to diable auto-eject */
	  if ( (ioctl(fp, CDROMEJECT_SW, 0)) == -1 )
        printf("... oops! error during CD-ROM init\n");
	  close(fp);
	} else {
	  printf("... oops! unable to open the CD-ROM device \"%s\" (CD-ROM init)\n",cdromdev);
	}
  } else {
	printf("no CD-ROM device configured! (CD-ROM init)\n");
  }
}

/* close or eject CD-ROM drive */
void DoEject(void) {
  int fp;
  /* do we have a CD-ROM device configured?  (extra check, not really nescessary) */
  if (strcmp(cdromdev,"")!=0) {
	/* try to open the device .. */
    if ( (fp = open(cdromdev, O_RDONLY | O_NONBLOCK)) != -1 ) {
	  if (!cdromstatus) { /* assumed closed */
		/* enable the drives software eject */
		if ( (ioctl(fp, CDROMEJECT_SW, 1)) == -1 )
          printf("... oops! error enabling CD-ROM SW eject\n");
		/* eject the cdrom tray */
		if (verbosemode) printf("... ejecting the CD-ROM tray\n");
        if ( (ioctl(fp, CDROMEJECT)) == -1 )
          printf("... oops! failed to eject the CD-ROM\n");
		else
		  cdromstatus = !cdromstatus;
	  } else { /* assumed open */
        /* disable the drives software eject again.. */
        if ( (ioctl(fp, CDROMEJECT_SW, 0)) == -1 )
          printf("... oops! error disabling CD-ROM SW eject\n");
		/* close the cdrom tray */
		if (verbosemode) printf("... closing the CD-ROM tray\n");
        if ( (ioctl(fp, CDROMCLOSETRAY)) == -1 )
          printf("... oops! failed to close the CD-ROM\n");
		else
		  cdromstatus = !cdromstatus;
	  }
      close(fp);
	} else {
	  printf("... oops! unable to open the CD-ROM device \"%s\"\n",cdromdev);
	}
  } else {
	printf("... oops! no CD-ROM device configured!\n");
  }
}

/**********************************************************************************************/

/* init sound: read and store the mixer master volume */
void initSound(void) {
  int fp;
  if (verbosemode) printf("Sound init, using \"%s\" as the mixer device\n",mixdev);
  if ( (fp = open( mixdev, O_RDWR|O_NONBLOCK )) != -1 ) {
    if ( SOUND_IOCTL(fp, SOUND_MIXER_READ_VOLUME, &old_mastervol) != -1) {
	  if (verbosemode) printf("... master volume stored\n");
	} else {
      printf("... oops! unable to read the volume of \"%s\" (sound init)\n", mixdev);
 	}
  } else {
    printf("... oops! unable to open the mixer device \"%s\" (sound init)\n", mixdev);
  }
}

/* adjust master volume or mute it */
/* return the value of the volume, or -1 if muted, -2 if not changed. */
int AdjustVolume(int value) {
  int fp, master_vol, left, right;
  int retval=0;
  /* open the mixer device */
  if ( (fp = open( mixdev, O_RDWR|O_NONBLOCK )) != -1 ) {
     if (value == VOLUP_VALUE || value == VOLDOWN_VALUE) { /* volume up/down */
        if (ismuted) { 
	   if (verbosemode) 
	      printf("... but we're muted\n"); 
	   close(fp); 
	   return -1; 
	}
	if ( SOUND_IOCTL(fp, SOUND_MIXER_READ_VOLUME, &master_vol) != -1) {
	   /* adjust the master volume */
           left = (master_vol & 0xFF) + value;
           right = ((master_vol >> 8) & 0xFF) + value;
           left = (left>MAX_VOLUME) ? MAX_VOLUME : ((left<0) ? 0 : left);
           right = (right>MAX_VOLUME) ? MAX_VOLUME : ((right<0) ? 0 : right);
           master_vol = left + (right << 8);
           if (SOUND_IOCTL(fp, SOUND_MIXER_WRITE_VOLUME, &master_vol) != -1) {
	      retval = master_vol;
              if (verbosemode) printf("... volume adjusted\n");
           } else {
              retval = -2;
	      printf("... oops! unable to adjust the master volume\n");
           }
        } else {
           printf("... oops! unable to read the volume of \"%s\"", mixdev);
 	}
     } else if (value == VOLMUTE_VALUE) { 
	if (!ismuted) { /* mute */
	   /* Store the old volume level */
	   if ( SOUND_IOCTL(fp, SOUND_MIXER_READ_VOLUME, &old_mastervol) != -1) {
	      if (verbosemode) printf("... old master volume stored\n");
	      /* Mute the volume */
              if (SOUND_IOCTL(fp, SOUND_MIXER_WRITE_VOLUME, &value) != -1) {
                 if (verbosemode) printf("... master volume muted\n");
		 ismuted = 1;
		 retval = -1;
              } else {
		 printf("... oops! unable to mute the master volume\n");
		 retval = -2;
              }
	   } else {
              printf("... oops! unable to read the volume of \"%s\"\n", mixdev);
	      retval = -2;
           }
	} else { /* un-mute */
           if (SOUND_IOCTL(fp, SOUND_MIXER_WRITE_VOLUME, &old_mastervol) != -1) {
              if (verbosemode) printf("... master volume restored\n");
	      ismuted = 0;
	      retval = old_mastervol;
           } else {
              printf("... oops! unable to restore the master volume\n");
	      retval = -2;
           }
	}
     }
     close(fp);
  } else {
    printf("... oops! unable to open the mixer device \"%s\"\n", mixdev);
    retval = -2;
  }
  return retval;
}

/**********************************************************************************************/

/* these functions provide for a somewhat cleaner exit */
static void cleanexit (int status) {
  cleanup();
  #ifdef HAVE_XOSD
  cleanup_lxosd(osd);
  #endif

  exit(status);
}

/* cleanup() only used when really bailing out :) */
static void cleanup (void) {
  /* free the dynamically assigned file names and devices*/
  if (displayname!=NULL) { free(displayname); displayname=NULL; }
  if (conffilename!=NULL) { free(conffilename); conffilename = NULL; }
  if (usrdeffilename!=NULL) { free(usrdeffilename); usrdeffilename = NULL; }
  if (deffilename!=NULL) { free(deffilename); deffilename = NULL; }
  if (kbtype!=NULL) { free(kbtype); kbtype=NULL; }
  if (cdromdev!=NULL) { free(cdromdev); cdromdev=NULL; }
  if (mixdev!=NULL) { free(mixdev); mixdev=NULL; }
  /* clean keyboard settings */
  if (myEAK.EAKbrand !=NULL) {
	free(myEAK.EAKbrand);
	myEAK.EAKbrand=NULL;
  }
  if (myEAK.EAKmodel != NULL) {
	free(myEAK.EAKmodel);
	myEAK.EAKmodel=NULL;
  }
  /* and clean up that linked list */
  cleanKeylist();
}

/* destroy the EAkeys linked list */
void cleanKeylist (void) {
  EAkey *tmp = NULL, *tmp2 = NULL;
  tmp = myEAK.EAKeylist;

  while (tmp!=NULL) {
	tmp2 = tmp->next;
	if (tmp->EAkeyname!=NULL) { free(tmp->EAkeyname); tmp->EAkeyname = NULL; }
	if (tmp->EAcommand!=NULL) { free(tmp->EAcommand); tmp->EAcommand = NULL; }
	free(tmp);
	tmp = tmp2;
  }
}

/* this is our "clean-up-that-mess" signal handler ;-) */
static void signalexit(int sig_num) {
  sigset_t mask_set; /* our signal masking set. */
  sigset_t old_set;  /* the old mask set.   */

  /* XOSD is multithreaded, and does not handle HUP or KILL signals gracefully.
   * how can we fix that? Can we put code here to help? */
  #ifdef HAVE_XOSD
  cleanup_lxosd(osd);
  #endif
    
  /* mask (block) any further signals while we're inside this signal handler. */
  sigfillset(&mask_set);
  sigprocmask(SIG_SETMASK, &mask_set, &old_set);
  /* now do some cleaning... */
  if (verbosemode) printf("cleaning up...\n");
  cleanupX();
  cleanupX();
  cleanup();

  if (verbosemode) printf("bye!\n");
  /* .. restore the old signal mask .. */
  sigprocmask(SIG_SETMASK, &old_set, NULL);
  
  /* and finally exit.. */
  exit(1);
}

/* HUP signal handler. reload .conf .def files and re-init program */
static void signalHUP(int sig_num) {
  sigset_t mask_set; /* our signal masking set. */
  sigset_t old_set;  /* the old mask set.   */
  /* mask (block) any further signals while we're inside this signal handler. */
  sigfillset(&mask_set);
  sigprocmask(SIG_SETMASK, &mask_set, &old_set);  
  printf("*** HUP signal caught. rehashing...\n");
  printf("*** Shutting down xosd safely...\n");
  #ifdef HAVE_XOSD
  cleanup_lxosd(osd);
  #endif
  /* clean keyboard settings */
  if (kbtype!=NULL) { free(kbtype); kbtype=NULL; }
  if (displayname!=NULL) { free(displayname); displayname=NULL; }
  if (myEAK.EAKbrand != NULL) { free(myEAK.EAKbrand); myEAK.EAKbrand=NULL; }
  if (myEAK.EAKmodel != NULL ) { free(myEAK.EAKmodel); myEAK.EAKmodel=NULL; }
  /* destroy EAkeyboard keylist */
  if (verbosemode) printf("*** cleaning up old keylist\n");
  cleanKeylist();
  /* clean devices */
  if (cdromdev!=NULL) { free(cdromdev); cdromdev=NULL; }
  if (mixdev !=NULL) { free(mixdev); mixdev=NULL; }
  /* reload .conf and .def file */
  if (verbosemode) printf("*** reloading .conf file \"%s\"\n",conffilename);
  parseconffile();
  if (verbosemode) printf("*** reloading .def file \"%s\"\n",deffilename);
  parsedeffile();
  /* re-init the EAKeyboard */
  if (verbosemode) printf("*** re-initializing the EAKeyboard\n");
  initEAK();
  /* destroy data structures */
  if (verbosemode) printf("*** cleaning up data structures\n");
  dcfp_destroy_data_struct((dcfp_data_struct *)confdata);
  dcfp_destroy_data_struct((dcfp_data_struct *)defdata);
  if (usrdefdata!=NULL) { free((dcfp_data_struct *)usrdefdata); usrdefdata = NULL; }
  /* re-init X/Xkb with our EAK keycodes/keysyms */
  if (verbosemode) printf("*** reconfiguring Xkb keymap with new keycodes/keysyms\n");
  initializeX();
  /* .. restore the old signal mask .. */
  if (verbosemode) printf("*** rehash done.\n");
  sigprocmask(SIG_SETMASK, &old_set, NULL);
  #ifdef HAVE_XOSD
  if (verbosemode) printf("*** Starting xosd...\n");
  osd = init_lxosd(font, color, delay, pos, offset, shadow, lines);
  #endif

  return;
}

/**********************************************************************************************/

/* X error handler. When a BadValue error occurs during XkbSetMap, we want to retry */
static int EAKXhandler(Display* theDisplay, XErrorEvent* xev) {
    // BadValue Error   (these numbers might not be platform or version independant?!!!):
    // xev->request_code = 150 (XKEYBOARD)
    // xev->minor_code   = 9   (XkbSetMap)
  if (nrretries < MAXRETRIES) {
	nrretries++;
    if (xev->request_code==150 && xev->minor_code==9)
      caughtXerror = 1;
  } else {
	/* max nr retries failed, quit to prevent race condition */
    printf("*** Xlib error caught, %d retries failed, exiting...\n",MAXRETRIES);
	cleanexit(1);
  }
  return 0;
}

/* dummy for SetAfterFunction */
static int dummyAfterFunc() { return 0; }

#ifdef HAVE_XOSD
/* unutuakuze xisd */
xosd* init_lxosd(char *font, char*color, int delay, xosd_pos pos, int offset, int shadow, int lines) {
    xosd *tosd;
    tosd = xosd_init (font, color, delay, pos, offset, shadow, lines);
    if (!tosd)
    {
        fprintf (stderr, "Error initializing osd\n");
        cleanexit(EXIT_FAILURE);
    }

   /* Not really needed, but at least we aren't throwing around an unknown value */
    old_age.tv_sec=0;

    if (scroll_age)
      gettimeofday(&old_age,0); 
    return tosd;
}

void cleanup_lxosd(xosd *tosd) {
  if (xosd_is_onscreen(tosd)) {
    xosd_wait_until_no_display(tosd);
  }
  xosd_uninit (tosd);
}

void lxosd_agenscroll(void) {
  /* Should we age the display? */
	
  if (scroll_age)
  {
     gettimeofday(&new_age,0);
     if ((new_age.tv_sec - old_age.tv_sec) > scroll_age)
     {
        xosd_scroll(osd,xosd_get_number_lines(osd));
        screen_line=0;
     } 
  }

  if (screen_line >= xosd_get_number_lines(osd))
  {
     xosd_scroll(osd,1);
     screen_line = xosd_get_number_lines(osd)-1;
  }
  if (forcewait && xosd_is_onscreen(osd)) {
     xosd_wait_until_no_display(osd);
  }

}

void lxosd_age(void) {
  old_age.tv_sec = new_age.tv_sec;
}

/* Modify the volume display output */
void lxosd_volume(int volume) {
  float volcalc;
  
  if (verbosemode) printf("volume is: %d\n", volume);
  if (volume != -2 ) {
     if (volume == -1) {
        xosd_display (osd, screen_line++, XOSD_string, "Volume Muted");
	lxosd_agenscroll();
     }
     else {
	volcalc = (float)volume;
	volcalc /= (float)max_audio;
	volcalc *= 100.00;
	volume = (int) volcalc;
        xosd_display (osd, 0, XOSD_string, "Volume");
        xosd_display (osd, 1, XOSD_percentage, volume);
	lxosd_agenscroll();
     }
  }
  
}
/* Display an onscreen command */
void lxosd_show(char *command) {
  xosd_display (osd, screen_line++, XOSD_string, command);
  lxosd_agenscroll();
}
#endif   

/* initialize X (xkb) so it understands the extra keys */
void initializeX(void) {
  XkbMessageAction xma;
  XkbMapChangesRec mapChangeRec;
  int types[1];
  EAkey *thisKey = NULL;

  ourdisplay = GetDisplay(displayname, &xkbOpcode, &xkbEventCode);
  if (!ourdisplay)
    cleanexit(1);

  /* catch Xlib error to retry */
  XSetAfterFunction(ourdisplay, dummyAfterFunc);
  XSetErrorHandler(EAKXhandler);

  /* construct the message action struct */
  xma.type = XkbSA_ActionMessage;
  xma.flags = XkbSA_MessageOnPress;
  strcpy(xma.message," ");

  /* add our new keysyms to the key codes of our display */
   for (thisKey = myEAK.EAKeylist; thisKey!=NULL;) {

    /* get the xkb map */
    xkb = XkbGetMap(ourdisplay, XkbAllMapComponentsMask, XkbUseCoreKbd);
    if (!xkb) {
      printf("FATAL ERROR: XkbGetMap failed\n");
  	  cleanexit(1);
    }

	/* check if our new keysym is valid */
	if (!XkbKeycodeInRange(xkb, thisKey->EAkeycode)) {
      printf("The keycode %d cannot be used, as it's not between the min(%d) and max(%d) keycode of your keyboard.\n"
             "Please increase the 'maximum' value in /usr/X11R6/lib/X11/xkb/keycodes/xfree86, then restart X.\n",
             thisKey->EAkeycode, xkb-> min_key_code, xkb->max_key_code);
      continue;
    }
    /* assign a group to the key code */
	types[0] = XkbOneLevelIndex;
    if ( XkbChangeTypesOfKey(xkb, thisKey->EAkeycode, 1, XkbGroup1Mask, types, NULL) != Success ) {
      printf("FATAL ERROR: XkbChangeTypesOfKey failed");
	  cleanexit(1);
    }
    /* change the keysym */
    if ( XkbResizeKeySyms( xkb, thisKey->EAkeycode, 1 ) == NULL ) {
      printf("FATAL ERROR: resize keysym failed\n");
	  cleanexit(1);
    }
	/* assign our new keysym to the keycode */
    *XkbKeySymsPtr(xkb,thisKey->EAkeycode) = thisKey->EAkeysym;
    /* add one key action to it */
    if ( XkbResizeKeyActions(xkb,thisKey->EAkeycode,1) == NULL ) {
      printf("FATAL ERROR: resize key action failed\n");
	  cleanexit(1);
    }
    commitXKBChanges(thisKey->EAkeycode);
    commitXKBChanges(thisKey->EAkeycode);
	  /* Anthony was right... this really needs to be called twice :)
	     When called only once, and the program is launched when logging into an X session,
	     one of every two keys is not functional ... (still weird thought that when launched
	     from an xterm, calling it only once _does_ work..)
	  */
    /* assign the message action to the key code */
    (&(xkb->server->acts[ xkb->server->key_acts[thisKey->EAkeycode] ]))[0] = (XkbAction) xma;
	/* commit the change back to the server. */
	bzero(&mapChangeRec, sizeof(mapChangeRec));
    mapChangeRec.changed = XkbKeyActionsMask;
    mapChangeRec.first_key_act = thisKey->EAkeycode;
    mapChangeRec.num_key_acts = 1;
    if ( XkbChangeMap(ourdisplay, xkb, &mapChangeRec) ) {
      if (verbosemode) printf("Xkb map change done for keycode %d\n",thisKey->EAkeycode);
    } else {
      printf("FATAL ERROR: map changed failed\n");
	  cleanexit(1);
    }
	if (!caughtXerror)
	  thisKey=thisKey->next;
	else {
	  if (verbosemode) printf("*** caught X error, retrying for keycode %d...\n",thisKey->EAkeycode);
	  caughtXerror = 0;
	}
  }

  /* Select the ActionMessage event under any circumstances */
  if ( !XkbSelectEvents(ourdisplay, XkbUseCoreKbd, XkbActionMessageMask, XkbActionMessageMask)) {
    printf("FATAL ERROR: Couldn't select desired XKB events\n");
    cleanexit(1);
  }
}

/* initialize X (xkb) so it understands the extra keys */
void cleanupX(void) {
  XkbMessageAction xma;
  XkbMapChangesRec mapChangeRec;
  /* int types[1]; */
  EAkey *thisKey = NULL;

  /* construct the message action struct */
  xma.type = KeyPress;
  xma.flags = XkbSA_MessageGenKeyEvent;
  strcpy(xma.message," ");

 /* get the xkb map */
  xkb = XkbGetMap(ourdisplay, XkbAllMapComponentsMask, XkbUseCoreKbd);
  if (!xkb) {
     printf("FATAL ERROR: XkbGetMap failed\n");
     cleanexit(1);
  }

  
  /* add our new keysyms to the key codes of our display */
   for (thisKey = myEAK.EAKeylist; thisKey!=NULL;) {

    /* change the keysym */
    if ( XkbResizeKeySyms( xkb, thisKey->EAkeycode, 0 ) == NULL ) {
          printf("FATAL ERROR: resize keysym failed\n");
	  cleanexit(1);
    }
    /* assign our new keysym to the keycode */
    commitXKBChanges(thisKey->EAkeycode);
    commitXKBChanges(thisKey->EAkeycode);
    /* assign the message action to the key code */
    (&(xkb->server->acts[ xkb->server->key_acts[thisKey->EAkeycode] ]))[0] = (XkbAction) xma;
    /* commit the change back to the server. */
    bzero(&mapChangeRec, sizeof(mapChangeRec));
    mapChangeRec.changed = XkbKeyActionsMask;
    mapChangeRec.first_key_act = thisKey->EAkeycode;
    mapChangeRec.num_key_acts = 0;
    if ( XkbChangeMap(ourdisplay, xkb, &mapChangeRec) ) {
      if (verbosemode) printf("Xkb map removed done for keycode %d\n",thisKey->EAkeycode);
    } else {
      printf("FATAL ERROR: map changed failed\n");
	  cleanexit(1);
    }
	if (!caughtXerror)
	  thisKey=thisKey->next;
	else {
	  if (verbosemode) printf("*** caught X error, retrying for keycode %d...\n",thisKey->EAkeycode);
	  caughtXerror = 0;
	} 
  }

  /* Select the ActionMessage event under any circumstances */
/*  if ( !XkbSelectEvents(ourdisplay, XkbUseCoreKbd, XkbAllEventsMask, XkbAllEventsMask)) {
    printf("FATAL ERROR: Couldn't select desired XKB events\n");
    cleanexit(1);
  } */
}

/**********************************************************************************************/

/* Get a hold on the display */
Display *GetDisplay(char* dpyName, int* opcodeRtrn, int* evBaseRtrn) {
  int mjr,mnr,error;
  Display *dpy;

  mjr = XkbMajorVersion;
  mnr = XkbMinorVersion;
  dpy = XkbOpenDisplay(dpyName,evBaseRtrn,NULL,&mjr,&mnr,&error);
  if (dpy == NULL) {
    switch (error) {
            case XkbOD_BadLibraryVersion:
                printf("%s was compiled with XKB version %d.%02d\n",
                        programname,XkbMajorVersion,XkbMinorVersion);
                printf("your X library supports incompatible version %d.%02d\n",
                        mjr,mnr);
                break;
            case XkbOD_ConnectionRefused:
                printf("Cannot open display \"%s\"\n",dpyName);
                break;
            case XkbOD_NonXkbServer:
                printf("XKB extension not present on %s\n",dpyName);
                break;
            case XkbOD_BadServerVersion:
                printf("%s was compiled with XKB version %d.%02d\n",
                        programname,XkbMajorVersion,XkbMinorVersion);
                printf("Server %s uses incompatible version %d.%02d\n",
                        dpyName,mjr,mnr);
                break;
            default:
                printf("Unknown error %d from XkbOpenDisplay\n",error);
      }
    } else {
      if (opcodeRtrn)
        XkbQueryExtension(dpy,opcodeRtrn,evBaseRtrn,NULL,&mjr,&mnr);
    }
    return dpy;
}

/**********************************************************************************************/

void commitXKBChanges(int tcode) {
    XkbMapChangesRec    mapChangeRec;

    /* Commit the change back to the server */
    bzero(&mapChangeRec, sizeof(mapChangeRec));
    mapChangeRec.changed = XkbKeySymsMask | XkbKeyTypesMask;
    mapChangeRec.first_key_sym = tcode;
    mapChangeRec.num_key_syms = 1;
    //        mapChangeRec.first_key_act = tcode;
    //        mapChangeRec.num_key_acts = 1;
    mapChangeRec.first_type = 0;
    mapChangeRec.num_types = xkb->map->num_types;
    if ( XkbChangeMap(ourdisplay, xkb, &mapChangeRec) ) {
      //printf("map change done: %d\n",tcode);
    } else {
      printf("FATAL ERROR: map changed failed\n");
	  cleanexit(1);
    }
}

/**********************************************************************************************/

/* show usage information on stdout */
void showusage(int argc, char *argv[]) {
  printf("\nLinEAK v" VERSION " -- Linux support for Easy Access and Internet Keyboards");
  printf("\n");
  printf("Usage: %s [options...]\n", argv[0]);
  printf("Valid options are:\n");
#ifdef HAVE_GETOPT_LONG
  printf("    -l,        --kbd-list           Show all supported keyboard types\n");
  printf("    -f FILE,   --conffile=FILE      Specify the .conf file\n");
  printf("    -e FILE,   --deffile=FILE       Specify the .def file\n");
  printf("    -d DEVICE, --cdrom-dev=DEVICE   Specify the CDROM device\n");
  printf("    -m DEVICE, --mixer-dev=DEVICE   Specify the mixer device\n");
  printf("    -c TYPE,   --config=TYPE        Create a new lineakd.conf file for keyboard TYPE\n"
         "                                      (warning: old one will be overwritten!)\n");
  printf("    -b,        --background         Run in background\n");
  printf("    -v,        --verbose            Print verbose messages\n"
         "                                      (only visible when not running in background)\n");
  printf("    -h,        --help               Print this message\n");
#else
  printf("    -l            Show all supported keyboard types\n");
  printf("    -f FILE       Specify the .conf file\n");
  printf("    -e FILE       Specify the .def file\n");
  printf("    -d DEVICE     Specify the CDROM device\n");
  printf("    -m DEVICE     Specify the mixer device\n");
  printf("    -c TYPE       Create a new lineakd.conf file for TYPE keyboard\n"
         "                    (warning: old one will be overwritten!)\n");
  printf("    -b            Run in background\n");
  printf("    -v,           Print verbose messages\n"
         "                    (only visible when not running in background)\n");
  printf("    -h            Print this message\n");
#endif
  printf("\nNOTE: if '-f', '-e', '-d', or '-m' parameters are used in conjunction with '-c', please\n      make sure you specify those BEFORE '-c' !\n");
  printf("\n");
}

/* parse incoming arguments */
int parseargs(int argc, char *argv[]) {
  int myopt, newconf=0;
  const char *flags = "hd:m:c:bvf:e:l";

#ifdef HAVE_GETOPT_LONG
  int option_index = 0;
  static struct option long_options[] = {
        {"help",            0, 0, 'h'},
        {"cdrom-dev",       1, 0, 'd'},
        {"mixer-dev",       1, 0, 'm'},
        {"create",          1, 0, 'c'},
        {"background",      0, 0, 'b'},
        {"verbose",         0, 0, 'v'},
        {"kbd-list",        0, 0, 'l'},
        {"conffile",        1, 0, 'f'},
        {"deffile",         1, 0, 'e'},
        {"verbose",         0, 0, 'v'},
        {0, 0, 0, 0}
  };

  while ((myopt = getopt_long(argc, argv, flags, long_options, &option_index)) != -1) {
#else
  while ((myopt = getopt(argc, argv, flags)) != -1) {
#endif

	/* parse each option */
    switch (myopt) {
          case 'd':
			  if (newconf) { // -d parameter has to appear BEFORE -c
				printf("*** -d parameter ignored, must be specified BEFORE -c\n");
				break;
			  }
              free(cdromdev);
			  cdromdev = malloc((strlen(optarg)+1)*sizeof(char));
		      sprintf(cdromdev, "%s", optarg);
              break;
          case 'm':
			  if (newconf) { // -m parameter has to appear BEFORE -c
				printf("*** -m parameter ignored, must be specified BEFORE -c\n");
				break;
			  }
              free(mixdev);
			  mixdev = malloc((strlen(optarg)+1)*sizeof(char));
	 	      sprintf(mixdev,"%s", optarg);
              break;
          case 'e':
			  if (newconf) { // -e parameter has to appear BEFORE -c
				printf("*** -e parameter ignored, must be specified BEFORE -c\n");
				break;
			  }
              free(deffilename);
			  deffilename = malloc((strlen(optarg)+1)*sizeof(char));
	 	      sprintf(deffilename,"%s", optarg);
              break;
          case 'f':
			  if (newconf) { // -f parameter has to appear BEFORE -c
				printf("*** -f parameter ignored, must be specified BEFORE -c\n");
				break;
			  }
              free(conffilename);
			  conffilename = malloc((strlen(optarg)+1)*sizeof(char));
	 	      sprintf(conffilename,"%s", optarg);
              break;
          case 'h':
              showusage(argc, argv);
              cleanexit(1);
              break;
          case 'c':
              printf("\nNOTE: if '-f', '-e', '-d', or '-m' parameters are used in conjunction with '-c', please\n      make sure you specify those BEFORE '-c' !\n\n");
		      kbtype = malloc((strlen(optarg)+1)*sizeof(char));
		      sprintf(kbtype, "%s", optarg);
			  parsedeffile();
              create_new_conf(NULL,optarg);
  		      newconf = 1;
              cleanexit(1);
              break;
          case 'b':
              runasdaemon = 1;
              break;
          case 'v':
              verbosemode = 1;
              break;
          case 'l':
			  parsedeffile();
              showkeyboards();
              cleanexit(1);
              break;
    }
  }
  
  /* set the program name without the directories */
  if ( strrchr( argv[0], '/' ) ) {
      programname = (char*) strrchr( argv[0], '/' ) + 1;
  } else {
      programname = argv[0];
  }

  /* check for a single additional argument */
  if ((argc - optind) > 1) {
    fprintf( stderr, "%s: too many arguments\n", argv[0] );
	cleanexit(1);
  }

  return 1;
}

/**********************************************************************************************/
